from typing import List, Optional, Tuple

import numpy as np
from numpy.random import RandomState

from alns.Outcome import Outcome
from alns.State import ContextualState
from alns.select.OperatorSelectionScheme import OperatorSelectionScheme

try:
    from mabwiser.mab import MAB, LearningPolicyType, NeighborhoodPolicyType

    MABWISER_AVAILABLE = True
except ModuleNotFoundError:
    MABWISER_AVAILABLE = False


class MABSelector(OperatorSelectionScheme):
    """
    A selector that uses any multi-armed-bandit algorithm from MABWiser.

    .. warning::

       ALNS does not install MABWiser by default. You can install it as an
       extra dependency via ``pip install alns[mabwiser]``.

    This selector is a wrapper around the many multi-armed bandit algorithms
    available in the `MABWiser <https://github.com/fidelity/mabwiser>`_
    library. Since ALNS operator selection can be framed as a
    multi-armed-bandit problem (where each [destroy, repair] operator pair is
    a bandit arm), this wrapper allows you to use a variety of existing
    multi-armed-bandit algorithms as operator selectors instead of
    having to reimplement them.

    .. note::

       If the provided learning policy is a contextual bandit algorithm, your
       state class must implement a ``get_context`` method that returns a
       context vector for the current state. See the
       :class:`~alns.State.ContextualState` protocol for details.

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the rewards when the
        candidate solution results in a new global best (idx 0), is better than
        the current solution (idx 1), the solution is accepted (idx 2), or
        rejected (idx 3).
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    learning_policy
        A MABWiser learning policy that acts as an operator selector. See the
        MABWiser documentation for a list of available learning policies.
    neighborhood_policy
        The neighborhood policy that MABWiser should use. Only available for
        contextual learning policies. See the MABWiser documentation for a
        list of available neighborhood policies.
    seed
        A seed that will be passed to the underlying MABWiser object.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    kwargs
        Any additional arguments. These will be passed to the underlying MAB
        object.

    References
    ----------
    .. [1] Emily Strong, Bernard Kleynhans, & Serdar Kadioglu (2021).
           MABWiser: Parallelizable Contextual Multi-armed Bandits.
           Int. J. Artif. Intell. Tools, 30(4), 2150021: 1 - 19.
    """

    def __init__(
        self,
        scores: List[float],
        num_destroy: int,
        num_repair: int,
        learning_policy: "LearningPolicyType",
        neighborhood_policy: Optional["NeighborhoodPolicyType"] = None,
        seed: Optional[int] = None,
        op_coupling: Optional[np.ndarray] = None,
        **kwargs,
    ):
        if not MABWISER_AVAILABLE:
            msg = """
            The MABSelector requires the MABWiser dependency to be installed.
            You can install it using `pip install alns[mabwiser]`.
            """
            raise ModuleNotFoundError(msg)

        super().__init__(num_destroy, num_repair, op_coupling)

        if any(score < 0 for score in scores):
            raise ValueError("Negative scores are not understood.")

        if len(scores) < 4:
            # More than four is OK because we only use the first four.
            raise ValueError(f"Expected four scores, found {len(scores)}")

        self._scores = scores

        if seed is not None:
            kwargs["seed"] = seed

        arms = [
            f"{d_idx}_{r_idx}"
            for d_idx in range(num_destroy)
            for r_idx in range(num_repair)
            if self._op_coupling[d_idx, r_idx]
        ]

        self._mab = MAB(arms, learning_policy, neighborhood_policy, **kwargs)

    @property
    def scores(self) -> List[float]:
        return self._scores

    @property
    def mab(self) -> "MAB":
        return self._mab

    def __call__(  # type: ignore[override]
        self,
        rnd_state: RandomState,
        best: ContextualState,
        curr: ContextualState,
    ) -> Tuple[int, int]:
        """
        Returns the (destroy, repair) operator pair from the underlying MAB
        strategy.
        """
        if not self._mab._is_initial_fit:  # noqa: SLF001
            # The MAB object has not yet been fit. In that case we return any
            # feasible operator index pair as a first observation.
            allowed = np.argwhere(self._op_coupling)
            idx = rnd_state.randint(len(allowed))
            return allowed[idx][0], allowed[idx][1]

        has_ctx = self._mab.is_contextual
        ctx = np.atleast_2d(curr.get_context()) if has_ctx else None
        prediction = self._mab.predict(contexts=ctx)
        return arm2ops(prediction)

    def update(  # type: ignore[override]
        self,
        cand: ContextualState,
        d_idx: int,
        r_idx: int,
        outcome: Outcome,
    ):
        """
        Updates the underlying MAB algorithm given the reward of the chosen
        destroy and repair operator combination ``(d_idx, r_idx)``.
        """
        has_ctx = self._mab.is_contextual
        ctx = np.atleast_2d(cand.get_context()) if has_ctx else None
        self._mab.partial_fit(
            [ops2arm(d_idx, r_idx)],
            [self._scores[outcome]],
            contexts=ctx,
        )


def ops2arm(d_idx: int, r_idx: int) -> str:
    """
    Converts the given destroy and repair operator indices to an arm string
    that can be passed to the MAB instance.

    Examples
    --------
    >>> ops2arm(0, 1)
    "0_1"
    >>> ops2arm(12, 3)
    "12_3"
    """
    return f"{d_idx}_{r_idx}"


def arm2ops(arm: str) -> Tuple[int, int]:
    """
    Converts an arm string returned by the MAB instance into a tuple of destroy
    and repair operator indices.

    Examples
    --------
    >>> arm2ops("0_1")
    (0, 1)
    >>> arm2ops("12_3")
    (12, 3)
    """
    d_idx, r_idx = map(int, arm.split("_"))
    return d_idx, r_idx



import logging
import time
from typing import Dict, List, Optional, Protocol, Tuple

import numpy.random as rnd

from alns.Outcome import Outcome
from alns.Result import Result
from alns.State import State
from alns.Statistics import Statistics
from alns.accept import AcceptanceCriterion
from alns.select import OperatorSelectionScheme
from alns.stop import StoppingCriterion


class _OperatorType(Protocol):
    __name__: str

    def __call__(
        self,
        state: State,
        rnd_state: rnd.RandomState,
        **kwargs,
    ) -> State:
        pass  # pragma: no cover


class _CallbackType(Protocol):
    __name__: str

    def __call__(self, state: State, rnd_state: rnd.RandomState, **kwargs):
        pass  # pragma: no cover


logger = logging.getLogger(__name__)


class ALNS:
    """
    Implements the adaptive large neighbourhood search (ALNS) algorithm.
    The implementation optimises for a minimisation problem, as explained
    in the text by Pisinger and Røpke (2010).

    .. note::

        Like the operators passed into the ALNS instance, any registered
        callback functions (registered via :meth:`~alns.ALNS.ALNS.on_best`,
        :meth:`~alns.ALNS.ALNS.on_better`, :meth:`~alns.ALNS.ALNS.on_accept`,
        or :meth:`~alns.ALNS.ALNS.on_reject`) should take a candidate
        :class:`~alns.State.State` and :class:`~numpy.random.RandomState` as
        arguments. Unlike the operators, no solution should be returned: if
        desired, the given candidate solution should be modified in-place
        instead. Note that this solution is **not** evaluated again (so a
        rejected candidate solution will stay rejected!).

    Parameters
    ----------
    rnd_state
        Optional random state to use for random number generation. When
        passed, this state is used for operator selection and general
        computations requiring random numbers. It is also passed to the
        destroy and repair operators, as a second argument.

    References
    ----------
    .. [1] Pisinger, D., and Røpke, S. (2010). Large Neighborhood Search. In
           M. Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399
           - 420). Springer.
    """

    def __init__(self, rnd_state: rnd.RandomState = rnd.RandomState()):
        self._rnd_state = rnd_state

        self._d_ops: Dict[str, _OperatorType] = {}
        self._r_ops: Dict[str, _OperatorType] = {}

        # Registers callback for each possible evaluation outcome.
        self._on_outcome: Dict[Outcome, _CallbackType] = {}

    @property
    def destroy_operators(self) -> List[Tuple[str, _OperatorType]]:
        """
        Returns the destroy operators set for the ALNS algorithm.

        Returns
        -------
        list
            A list of (name, operator) tuples. Their order is the same as the
            one in which they were passed to the ALNS instance.
        """
        return list(self._d_ops.items())

    @property
    def repair_operators(self) -> List[Tuple[str, _OperatorType]]:
        """
        Returns the repair operators set for the ALNS algorithm.

        Returns
        -------
        list
            A list of (name, operator) tuples. Their order is the same as the
            one in which they were passed to the ALNS instance.
        """
        return list(self._r_ops.items())

    def add_destroy_operator(
        self, op: _OperatorType, name: Optional[str] = None
    ):
        """
        Adds a destroy operator to the heuristic instance.

        .. warning::

            A destroy operator will receive the current solution state
            maintained by the ALNS instance, not a copy. Make sure to modify
            a **copy** of this state in the destroy operator, created using,
            for example, :func:`copy.copy` or :func:`copy.deepcopy`.

        Parameters
        ----------
        op
            An operator that, when applied to the current state, returns a new
            state reflecting its implemented destroy action. Its second
            argument is the random state passed to the ALNS instance.
        name
            Optional name argument, naming the operator. When not passed, the
            function name is used instead.
        """
        logger.debug(f"Adding destroy operator {op.__name__}.")
        self._d_ops[op.__name__ if name is None else name] = op

    def add_repair_operator(
        self, op: _OperatorType, name: Optional[str] = None
    ):
        """
        Adds a repair operator to the heuristic instance.

        Parameters
        ----------
        op
            An operator that, when applied to the destroyed state, returns a
            new state reflecting its implemented repair action. Its second
            argument is the random state passed to the ALNS instance.
        name
            Optional name argument, naming the operator. When not passed, the
            function name is used instead.
        """
        logger.debug(f"Adding repair operator {op.__name__}.")
        self._r_ops[name if name else op.__name__] = op

    def iterate(
        self,
        initial_solution: State,
        op_select: OperatorSelectionScheme,
        accept: AcceptanceCriterion,
        stop: StoppingCriterion,
        **kwargs,
    ) -> Result:
        """
        Runs the adaptive large neighbourhood search heuristic [1], using the
        previously set destroy and repair operators. The first solution is set
        to the passed-in initial solution, and then subsequent solutions are
        computed by iteratively applying the operators.

        Parameters
        ----------
        initial_solution
            The initial solution, as a State object.
        op_select
            The operator selection scheme to use for selecting operators.
            See also the ``alns.select`` module for an overview.
        accept
            The acceptance criterion to use for candidate states.
            See also the ``alns.accept`` module for an overview.
        stop
            The stopping criterion to use for stopping the iterations.
            See also the ``alns.stop`` module for an overview.
        **kwargs
            Optional keyword arguments. These are passed to the operators and
            any registered callbacks.

        Raises
        ------
        ValueError
            When the parameters do not meet requirements.

        Returns
        -------
        Result
            A result object, containing the best solution and some additional
            statistics.

        References
        ----------
        .. [1] Pisinger, D., & Røpke, S. (2010). Large Neighborhood Search. In
               M. Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399
               - 420). Springer.

        .. [2] S. Røpke and D. Pisinger (2006). A unified heuristic for a large
               class of vehicle routing problems with backhauls. *European
               Journal of Operational Research*, 171: 750-775.
        """
        if len(self.destroy_operators) == 0 or len(self.repair_operators) == 0:
            raise ValueError("Missing destroy or repair operators.")

        curr = best = initial_solution
        init_obj = initial_solution.objective()

        logger.debug(f"Initial solution has objective {init_obj:.2f}.")

        stats = Statistics()
        stats.collect_objective(init_obj)
        stats.collect_runtime(time.perf_counter())

        while not stop(self._rnd_state, best, curr):
            d_idx, r_idx = op_select(self._rnd_state, best, curr)

            d_name, d_operator = self.destroy_operators[d_idx]
            r_name, r_operator = self.repair_operators[r_idx]

            logger.debug(f"Selected operators {d_name} and {r_name}.")

            destroyed = d_operator(curr, self._rnd_state, **kwargs)
            cand = r_operator(destroyed, self._rnd_state, **kwargs)

            best, curr, outcome = self._eval_cand(
                accept, best, curr, cand, **kwargs
            )

            op_select.update(cand, d_idx, r_idx, outcome)

            stats.collect_objective(curr.objective())
            stats.collect_destroy_operator(d_name, outcome)
            stats.collect_repair_operator(r_name, outcome)
            stats.collect_runtime(time.perf_counter())

        logger.info(f"Finished iterating in {stats.total_runtime:.2f}s.")

        return Result(best, stats)

    def on_best(self, func: _CallbackType):
        """
        Sets a callback function to be called when ALNS finds a new global best
        solution state.
        """
        logger.debug(f"Adding on_best callback {func.__name__}.")
        self._on_outcome[Outcome.BEST] = func

    def on_better(self, func: _CallbackType):
        """
        Sets a callback function to be called when ALNS finds a better solution
        than the current incumbent.
        """
        logger.debug(f"Adding on_better callback {func.__name__}.")
        self._on_outcome[Outcome.BETTER] = func

    def on_accept(self, func: _CallbackType):
        """
        Sets a callback function to be called when ALNS accepts a new solution
        as the current incumbent (that is not a new global best, or otherwise
        improving).
        """
        logger.debug(f"Adding on_accept callback {func.__name__}.")
        self._on_outcome[Outcome.ACCEPT] = func

    def on_reject(self, func: _CallbackType):
        """
        Sets a callback function to be called when ALNS rejects a new solution.
        """
        logger.debug(f"Adding on_reject callback {func.__name__}.")
        self._on_outcome[Outcome.REJECT] = func

    def _eval_cand(
        self,
        accept: AcceptanceCriterion,
        best: State,
        curr: State,
        cand: State,
        **kwargs,
    ) -> Tuple[State, State, Outcome]:
        """
        Considers the candidate solution by comparing it against the best and
        current solutions. Candidate solutions are accepted based on the
        passed-in acceptance criterion. The (possibly new) best and current
        solutions are returned, along with a weight index (best, better,
        accepted, rejected).

        Returns
        -------
        tuple
            A tuple of the best and current solution, along with the weight
            index.
        """
        outcome = self._determine_outcome(accept, best, curr, cand)
        func = self._on_outcome.get(outcome)

        if callable(func):
            func(cand, self._rnd_state, **kwargs)

        if outcome == Outcome.BEST:
            return cand, cand, outcome

        if outcome == Outcome.REJECT:
            return best, curr, outcome

        return best, cand, outcome

    def _determine_outcome(
        self,
        accept: AcceptanceCriterion,
        best: State,
        curr: State,
        cand: State,
    ) -> Outcome:
        """
        Determines the candidate solution's evaluation outcome.
        """
        outcome = Outcome.REJECT

        if accept(self._rnd_state, best, curr, cand):  # accept candidate
            outcome = Outcome.ACCEPT

            if cand.objective() < curr.objective():
                outcome = Outcome.BETTER

        if cand.objective() < best.objective():  # candidate is new best
            logger.info(f"New best with objective {cand.objective():.2f}.")
            outcome = Outcome.BEST

        return outcome



from collections import deque
from statistics import mean
from typing import Deque, List


class MovingAverageThreshold:
    """
    The Moving Average Threshold (MAT) criterion of [1]. This criterion accepts
    a candidate solution if it is better than a threshold value that is based
    on the moving average of the objective values of recently observed
    candidate solutions. The threshold is computed as:

    .. math::

       f(s^*) + \\eta \\left(
          \\sum_{i = 1}^\\gamma \\frac{f(s^i)}{\\gamma} - f(s^*)
       \\right)

    where :math:`s^*` is the best solution observed in the last :math:`\\gamma`
    iterations, :math:`f(\\cdot)` indicates the objective function,
    :math:`\\eta \\in [0, 1]` and :math:`\\gamma \\in \\mathbb{N}` are
    parameters, and each :math:`s^i` is a recently observed solution. The
    recently observed solutions are stored in a ``history`` attributed of size
    at most :math:`\\gamma`.

    Parameters
    ----------
    eta: float
        Used to determine the threshold value. Larger values of :math:`\\eta`
        result in more accepted candidate solutions. Must be in [0, 1].
    gamma: int
        History size. Must be positive.

    References
    ----------
    .. [1] Máximo, V.R. and M.C.V. Nascimento. 2021. A hybrid adaptive iterated
           local search with diversification control to the capacitated vehicle
           routing problem, *European Journal of Operational Research* 294 (3):
           1108 - 1119.
    """

    def __init__(self, eta: float, gamma: int):
        if not (0 <= eta <= 1):
            raise ValueError("eta must be in [0, 1].")

        if gamma <= 0:
            raise ValueError("gamma must be positive.")

        self._eta = eta
        self._gamma = gamma
        self._history: Deque[float] = deque(maxlen=gamma)

    @property
    def eta(self) -> float:
        return self._eta

    @property
    def gamma(self) -> int:
        return self._gamma

    @property
    def history(self) -> List[float]:
        return list(self._history)

    def __call__(self, rnd, best, current, candidate) -> bool:
        self._history.append(candidate.objective())
        recent_best = min(self._history)
        recent_avg = mean(self._history)

        threshold = recent_best + self._eta * (recent_avg - recent_best)
        return candidate.objective() <= threshold



from .AcceptanceCriterion import AcceptanceCriterion
from .AlwaysAccept import AlwaysAccept
from .GreatDeluge import GreatDeluge
from .HillClimbing import HillClimbing
from .LateAcceptanceHillClimbing import LateAcceptanceHillClimbing
from .MovingAverageThreshold import MovingAverageThreshold
from .NonLinearGreatDeluge import NonLinearGreatDeluge
from .RandomAccept import RandomAccept
from .RecordToRecordTravel import RecordToRecordTravel
from .SimulatedAnnealing import SimulatedAnnealing



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np
from numpy.random import RandomState

from alns.Outcome import Outcome
from alns.State import State


class OperatorSelectionScheme(ABC):
    """
    Base class describing an operator selection scheme.

    Parameters
    ----------
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional 2D boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    """

    def __init__(
        self,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        if op_coupling is not None:
            op_coupling = np.asarray(op_coupling, dtype=bool)
            op_coupling = np.atleast_2d(op_coupling)
        else:
            op_coupling = np.ones((num_destroy, num_repair), dtype=bool)

        self._validate_arguments(num_destroy, num_repair, op_coupling)

        self._num_destroy = num_destroy
        self._num_repair = num_repair
        self._op_coupling = op_coupling

    @property
    def num_destroy(self) -> int:
        return self._num_destroy

    @property
    def num_repair(self) -> int:
        return self._num_repair

    @property
    def op_coupling(self) -> np.ndarray:
        return self._op_coupling

    @abstractmethod
    def __call__(
        self, rnd: RandomState, best: State, curr: State
    ) -> Tuple[int, int]:
        """
        Determine which destroy and repair operator pair to apply in this
        iteration.

        Parameters
        ----------
        rnd_state
            Random state object, to be used for random number generation.
        best
            The best solution state observed so far.
        current
            The current solution state.

        Returns
        -------
        A tuple of (d_idx, r_idx), which are indices into the destroy and
        repair operator lists, respectively.
        """
        return NotImplemented

    @abstractmethod
    def update(
        self, candidate: State, d_idx: int, r_idx: int, outcome: Outcome
    ):
        """
        Updates the selection schame based on the outcome of the applied
        destroy (d_idx) and repair (r_idx) operators.

        Parameters
        ----------
        candidate
            The candidate solution state.
        d_idx
            Destroy operator index.
        r_idx
            Repair operator index.
        outcome
            Score enum value used for the various iteration outcomes.
        """
        return NotImplemented

    @staticmethod
    def _validate_arguments(
        num_destroy: int, num_repair: int, op_coupling: np.ndarray
    ):
        if num_destroy <= 0 or num_repair <= 0:
            raise ValueError("Missing destroy or repair operators.")

        if op_coupling.shape != (num_destroy, num_repair):
            raise ValueError(
                f"Coupling matrix of shape {op_coupling.shape}, expected "
                f"{(num_destroy, num_repair)}."
            )

        # Destroy ops. must be coupled with at least one repair operator
        d_idcs = np.flatnonzero(np.count_nonzero(op_coupling, axis=1) == 0)

        if d_idcs.size != 0:
            d_op = f"Destroy op. {d_idcs[0]}"
            raise ValueError(f"{d_op} has no coupled repair operators.")



from typing import List

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises
from pytest import mark

from alns.select import SegmentedRouletteWheel
from alns.tests.states import Zero


@mark.parametrize(
    "scores, decay, num_destroy, num_repair, seg_length, op_coupling",
    [
        ([0, 0, 0, 0], 0, 1, 1, 1, np.ones((1, 1))),
        ([0, 1, 2, 3], 0.2, 2, 2, 2, np.ones((2, 2))),
        ([5, 3, 2, 1], 1, 10, 10, 10, np.ones((10, 10))),
    ],
)
def test_properties(
    scores, decay, num_destroy, num_repair, seg_length, op_coupling
):
    select = SegmentedRouletteWheel(
        scores,
        decay,
        seg_length,
        num_destroy,
        num_repair,
        op_coupling=op_coupling,
    )

    assert_equal(select.seg_length, seg_length)


@mark.parametrize("decay", [1.01, -0.01, -0.5, 1.5])
def test_raises_invalid_decay(decay: float):
    with assert_raises(ValueError):
        SegmentedRouletteWheel([0, 0, 0, 0], decay, 100, 1, 1)


@mark.parametrize("decay", np.linspace(0, 1, num=5))
def test_does_not_raise_valid_decay(decay: float):
    SegmentedRouletteWheel([0, 0, 0, 0], decay, 100, 1, 1)


@mark.parametrize(
    "scores,decay,expected",
    [
        ([5, 3, 2, 1], 0, [3, 3]),  # 1 * 0 + (0 + 3) * 1 = 3
        ([5, 3, 2, 1], 1, [1, 1]),  # 1 * 1 + (0 + 3) * 0 = 1
        ([5, 3, 2, 1], 0.5, [2, 2]),  # .5 * 1 + (0 + 3) * .5 = 2
        ([5, 5, 5, 5], 0, [5, 5]),  # etc. etc.
        ([5, 5, 5, 5], 1, [1, 1]),
        ([5, 5, 5, 5], 0.5, [3, 3]),
    ],
)
def test_update(scores: List[float], decay: float, expected: List[float]):
    rnd_state = np.random.RandomState(1)
    select = SegmentedRouletteWheel(scores, decay, 1, 1, 1)

    # TODO other weights?
    select.update(Zero(), 0, 0, 1)
    select(rnd_state, Zero(), Zero())

    assert_almost_equal(select.destroy_weights[0], expected[0])
    assert_almost_equal(select.repair_weights[0], expected[1])


@mark.parametrize(
    "scores,num_destroy,num_repair,decay,seg_length",
    [
        ([5, 3, 2, -1], 1, 1, 0, 1),  # negative score
        ([5, 3, 2], 1, 1, 0, 1),  # len(score) < 4
        ([5, 3, 2, 1], 0, 1, 0, 1),  # no destroy operator
        ([5, 3, 2, 1], 1, 0, 0, 1),  # no repair operator
        ([5, 3, 2, 1], 1, 1, -1, 1),  # decay < 0
        ([5, 3, 2, 1], 1, 1, 2, 1),  # decay > 1
        ([5, 3, 2, 1], 1, 1, 0.5, 0),  # seg_length < 1
    ],
)
def test_raises_invalid_arguments(
    scores: List[float],
    num_destroy: int,
    num_repair: int,
    decay: float,
    seg_length: int,
):
    with assert_raises(ValueError):
        SegmentedRouletteWheel(
            scores, decay, seg_length, num_destroy, num_repair
        )


# TODO test select weights, at iteration start



import logging
from typing import List, Optional

import numpy as np

from alns.State import State
from alns.select.RouletteWheel import RouletteWheel

logger = logging.getLogger(__name__)


class SegmentedRouletteWheel(RouletteWheel):
    R"""
    .. note::

        First read the documentation for
        :class:`~alns.select.RouletteWheel.RouletteWheel`, the parent of this
        class.

    The ``SegmentedRouletteWheel`` scheme extends the
    :class:`~alns.select.RouletteWheel.RouletteWheel` scheme by fixing operator
    weights for a number of iterations (the *segment length*). This allows
    certain sets of operators to be selected more often in different
    neighbourhoods.

    Initially, all weights are set to one, as in
    :class:`~alns.select.RouletteWheel.RouletteWheel`.
    A separate score is tracked for each operator :math:`d` and :math:`r`, to
    which the observed scores :math:`s_j` are added in each iteration where
    :math:`d` and :math:`r` are applied. After the segment concludes, these
    summed scores are added to the existing weights :math:`\omega_d` and
    :math:`\omega_r` as a convex combination using a parameter :math:`\theta`
    (the *segment decay rate*), as in ``RouletteWheel``. The separate
    score list is then reset to zero, and a new segment begins.

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the weight
        updates when the candidate solution results in a new global best
        (idx 0), is better than the current solution (idx 1), the solution
        is accepted (idx 2), or rejected (idx 3).
    decay
        Operator decay parameter :math:`\theta \in [0, 1]`. This parameter is
        used to weigh the running performance of each operator.
    seg_length
        Length of a single segment.
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    """

    def __init__(
        self,
        scores: List[float],
        decay: float,
        seg_length: int,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        super().__init__(scores, decay, num_destroy, num_repair, op_coupling)

        if seg_length < 1:
            raise ValueError("seg_length < 1 not understood.")

        self._seg_length = seg_length
        self._iter = 0

        self._reset_segment_weights()

    @property
    def seg_length(self):
        return self._seg_length

    def __call__(self, rnd_state, best: State, curr: State):
        self._iter += 1

        if self._iter % self._seg_length == 0:
            logger.debug(f"End of segment (#iters = {self._iter}).")

            self._d_weights *= self._decay
            self._d_weights += (1 - self._decay) * self._d_seg_weights

            self._r_weights *= self._decay
            self._r_weights += (1 - self._decay) * self._r_seg_weights

            self._reset_segment_weights()

        return super().__call__(rnd_state, best, curr)

    def update(self, cand, d_idx, r_idx, outcome):
        self._d_seg_weights[d_idx] += self._scores[outcome]
        self._r_seg_weights[r_idx] += self._scores[outcome]

    def _reset_segment_weights(self):
        self._d_seg_weights = np.zeros_like(self._d_weights)
        self._r_seg_weights = np.zeros_like(self._r_weights)



from alns.accept.update import update


class RandomAccept:
    """
    The Random Accept criterion accepts a candidate solution if it improves
    over the current one, or with a given probability :math:`P` regardless of
    the cost. :math:`P` is updated in each iteration as:

    .. math::

        P \\gets \\max \\{ P_\\text{end},~P - \\gamma \\}

    when ``method = 'linear'``, or

    .. math::

        P \\gets \\max \\{ P_\\text{end},~\\gamma P \\}

    when ``method = 'exponential'``. Initially, :math:`P` is set to
    :math:`P_\\text{start}`.

    Parameters
    ----------
    start_prob
        The initial probability :math:`P_\\text{start} \\in [0, 1]`.
    end_prob
        The final probability :math:`P_\\text{end} \\in [0, 1]`.
    step:
        The updating step :math:`\\gamma \\ge 0`.
    method
        The updating method, one of {'linear', 'exponential'}. Default
        'linear'.
    """

    def __init__(
        self,
        start_prob: float,
        end_prob: float,
        step: float,
        method: str = "linear",
    ):
        if not 0 <= end_prob <= start_prob <= 1:
            raise ValueError("Must have 0 <= start_prob <= end_prob <= 1")

        if step < 0:
            raise ValueError("Step cannot be negative.")

        if method == "exponential" and step > 1:
            raise ValueError("Exponential updating cannot have step > 1.")

        self._start_prob = start_prob
        self._end_prob = end_prob
        self._step = step
        self._method = method

        self._prob = start_prob

    @property
    def start_prob(self) -> float:
        return self._start_prob

    @property
    def end_prob(self) -> float:
        return self._end_prob

    @property
    def step(self) -> float:
        return self._step

    @property
    def method(self) -> str:
        return self._method

    def __call__(self, rnd, best, current, candidate):
        # Always accept better
        res = candidate.objective() < current.objective()

        if not res:  # maybe accept worse
            res = rnd.random() < self._prob

        self._prob = max(
            self.end_prob, update(self._prob, self.step, self.method)
        )

        return res



from typing import Protocol

import numpy as np


class State(Protocol):
    """
    Protocol for a solution state. Solutions should define an ``objective()``
    member function for evaluation.
    """

    def objective(self) -> float:
        """
        Computes the state's associated objective value.
        """


class ContextualState(State, Protocol):
    """
    Protocol for a solution state that also provides context. Solutions should
    define ``objective()`` and ``get_context()`` methods.
    """

    def get_context(self) -> np.ndarray:
        """
        Computes a context vector for the current state.
        """



from typing import List

import numpy as np
import numpy.random as rnd
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_equal,
    assert_raises,
)
from pytest import mark

from alns.select import RouletteWheel
from alns.tests.states import Zero


@mark.parametrize(
    "scores, decay, num_destroy, num_repair, op_coupling",
    [
        ([0, 0, 0, 0], 0, 1, 1, np.ones((1, 1))),
        ([0, 1, 2, 3], 0.2, 2, 2, np.ones((2, 2))),
        ([5, 3, 2, 1], 1, 10, 10, np.ones((10, 10))),
    ],
)
def test_properties(scores, decay, num_destroy, num_repair, op_coupling):
    select = RouletteWheel(
        scores, decay, num_destroy, num_repair, op_coupling=op_coupling
    )

    # TODO move these property tests to RandomSelect
    assert_equal(select.num_destroy, num_destroy)
    assert_equal(select.num_repair, num_repair)
    assert_equal(select.op_coupling, op_coupling)

    assert_equal(select.scores, scores)
    assert_equal(select.destroy_weights, np.ones(num_destroy))
    assert_equal(select.repair_weights, np.ones(num_repair))
    assert_equal(select.decay, decay)


@mark.parametrize("decay", [1.01, -0.01, -0.5, 1.5])
def test_raises_invalid_decay(decay: float):
    with assert_raises(ValueError):
        RouletteWheel([0, 0, 0, 0], decay, 1, 1)


@mark.parametrize("decay", np.linspace(0, 1, num=5))
def test_does_not_raise_valid_decay(decay: float):
    RouletteWheel([0, 0, 0, 0], decay, 1, 1)


@mark.parametrize(
    "scores,decay,expected",
    [
        ([0, 0, 0, 0], 1, [1, 1]),  # scores are not used
        ([0, 0, 0, 0], 0, [0, 0]),  # initial weights are not used
        ([0.5, 0.5, 0.5, 0.5], 0.5, [0.75, 0.75]),
    ],
)  # convex combination
def test_update(scores: List[float], decay: float, expected: List[float]):
    select = RouletteWheel(scores, decay, 1, 1)

    # TODO other weights?
    select.update(Zero(), 0, 0, 1)

    assert_almost_equal(select.destroy_weights[0], expected[0])
    assert_almost_equal(select.repair_weights[0], expected[1])


@mark.parametrize(
    "op_coupling",
    [
        np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]),
        np.array([[1, 0, 1], [0, 1, 1], [1, 1, 0]]),
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]),
        np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]]),
    ],
)
def test_select_coupled_operators(op_coupling):
    """
    Test if the indices of the selected operators correspond to the
    ones that are given by the operator coupling.
    """
    rnd_state = rnd.RandomState()
    n_destroy, n_repair = op_coupling.shape
    select = RouletteWheel(
        [0, 0, 0, 0], 0, n_destroy, n_repair, op_coupling=op_coupling
    )
    d_idx, r_idx = select(rnd_state, Zero(), Zero())

    assert_((d_idx, r_idx) in np.argwhere(op_coupling == 1))


@mark.parametrize(
    "op_coupling", [np.zeros((2, 2)), np.array([[0, 0, 0], [1, 1, 1]])]
)
def test_raise_uncoupled_destroy_op(op_coupling):
    """
    Tests if having a destroy operator that is not coupled to any of the
    repair operators raises an an error.
    """
    with assert_raises(ValueError):
        n_destroy, n_repair = op_coupling.shape
        RouletteWheel(
            [0, 0, 0, 0], 0, n_destroy, n_repair, op_coupling=op_coupling
        )


@mark.parametrize(
    "n_destroy, n_repair, op_coupling",
    [
        (1, 2, [0]),  # missing repair column
        (2, 2, [0, 0]),  # missing destroy row
        (2, 1, [0, 0]),  # too many repair, too few destroy
    ],
)
def test_raises_wrong_op_coupling_shape(n_destroy, n_repair, op_coupling):
    with assert_raises(ValueError):
        RouletteWheel([0, 0, 0, 0], 0, n_destroy, n_repair, op_coupling)


def test_single_destroy_operator_coerces_coupling_matrix():
    select = RouletteWheel([0, 0, 0, 0], 0, 1, 2, [1, 0])
    assert_equal(select.op_coupling.shape, (1, 2))



import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.accept import MovingAverageThreshold
from alns.tests.states import One, Two, VarObj, Zero


@mark.parametrize(
    "eta, gamma",
    [
        (-1, 3),  # eta cannot be < 0
        (2, 3),  # eta cannot be > 1
        (0.5, -2),  # gamma cannot be < 0
        (0.5, 0),  # gamma cannot be 0
    ],
)
def test_raise_invalid_parameters(eta, gamma):
    with assert_raises(ValueError):
        MovingAverageThreshold(eta=eta, gamma=gamma)


@mark.parametrize("eta, gamma", [(1, 3), (0.4, 4)])
def test_no_raise_valid_parameters(eta, gamma):
    MovingAverageThreshold(eta=eta, gamma=gamma)


@mark.parametrize("eta", [0, 0.01, 0.5, 0.99, 1])
def test_eta(eta):
    moving_average = MovingAverageThreshold(eta, 3)
    assert_equal(moving_average.eta, eta)


@mark.parametrize("gamma", range(1, 10))
def test_gamma(gamma):
    moving_average = MovingAverageThreshold(0.5, gamma)
    assert_equal(moving_average.gamma, gamma)


def test_accepts_below_threshold():
    moving_average = MovingAverageThreshold(eta=0.5, gamma=4)
    moving_average(rnd.RandomState(), One(), One(), One())
    moving_average(rnd.RandomState(), One(), One(), Zero())

    # The threshold is set at 0 + 0.5 * (0.5 - 0) = 0.25
    assert_(moving_average(rnd.RandomState(), One(), One(), Zero()))


def test_rejects_above_threshold():
    moving_average = MovingAverageThreshold(eta=0.5, gamma=4)
    moving_average(rnd.RandomState(), One(), One(), Two())
    moving_average(rnd.RandomState(), One(), One(), Zero())

    # The threshold is set at 0 + 0.5 * (1 - 0) = 0.5
    assert_(not moving_average(rnd.RandomState(), One(), One(), One()))


def test_accepts_equal_threshold():
    moving_average = MovingAverageThreshold(eta=0.5, gamma=4)
    moving_average(rnd.RandomState(), One(), One(), VarObj(7100))
    moving_average(rnd.RandomState(), One(), One(), VarObj(7200))

    # The threshold is set at 7100 + 0.5 * (7140 - 7100) = 7120
    assert_(moving_average(rnd.RandomState(), One(), One(), VarObj(7120)))


def test_accepts_over_gamma_candidates():
    moving_average = MovingAverageThreshold(eta=0.2, gamma=3)
    moving_average(rnd.RandomState(), One(), One(), VarObj(7100))
    moving_average(rnd.RandomState(), One(), One(), VarObj(7200))
    moving_average(rnd.RandomState(), One(), One(), VarObj(7200))

    # The threshold is set at 7000 + 0.2 * (7133.33 - 7000) = 7013.33
    assert_(moving_average(rnd.RandomState(), One(), One(), VarObj(7000)))


def test_rejects_over_gamma_candidates():
    moving_average = MovingAverageThreshold(eta=0.2, gamma=3)

    for value in [7100, 7200, 7200, 7000]:
        moving_average(rnd.RandomState(), One(), One(), VarObj(value))

    # The threshold is set at 7000 + 0.2 * (7100 - 7000) = 7020
    result = moving_average(rnd.RandomState(), One(), One(), VarObj(7100))
    assert_(not result)


def test_evaluate_consecutive_solutions():
    """
    Test if MAT correctly accepts and rejects consecutive solutions.
    """
    moving_average = MovingAverageThreshold(eta=0.5, gamma=4)

    # The threshold is set at 7100, hence the solution is accepted.
    assert_(moving_average(rnd.RandomState(), One(), One(), VarObj(7100)))

    # The threshold is set at 7125, hence the solution is accepted.
    result = moving_average(rnd.RandomState(), One(), One(), VarObj(7200))
    assert_(not result)

    # The threshold is set at 7120, hence the solution is accepted.
    assert_(moving_average(rnd.RandomState(), One(), One(), VarObj(7120)))


def test_history():
    """
    Test if MAT correctly stores the history of the thresholds correctly.
    """
    moving_average = MovingAverageThreshold(eta=0.5, gamma=4)

    moving_average(rnd.RandomState(), One(), One(), VarObj(7100))
    assert_equal(moving_average.history, [7100])

    moving_average(rnd.RandomState(), One(), One(), VarObj(7200))
    assert_equal(moving_average.history, [7100, 7200])

    moving_average(rnd.RandomState(), One(), One(), VarObj(7120))
    assert_equal(moving_average.history, [7100, 7200, 7120])

    moving_average(rnd.RandomState(), One(), One(), VarObj(7100))
    assert_equal(moving_average.history, [7100, 7200, 7120, 7100])

    moving_average(rnd.RandomState(), One(), One(), VarObj(7200))
    assert_equal(moving_average.history, [7200, 7120, 7100, 7200])



from typing import List, Optional, Tuple

import numpy as np
from numpy.random import RandomState

from alns.State import State
from alns.select.OperatorSelectionScheme import OperatorSelectionScheme


class RouletteWheel(OperatorSelectionScheme):
    R"""
    The ``RouletteWheel`` scheme updates operator weights as a convex
    combination of the current weight, and the new score.

    When the algorithm starts, all operators :math:`i` are assigned weight
    :math:`\omega_i = 1`. In each iteration, a destroy and a repair operator
    are selected by the ALNS algorithm, based on the normalised current weights
    :math:`\omega_i`. The selected operators are applied to the current
    solution, resulting in a new candidate solution. This candidate is
    evaluated by the ALNS algorithm, which leads to one of four outcomes:

    1. The candidate solution is a new global best.
    2. The candidate solution is better than the current solution, but not a
       new global best.
    3. The candidate solution is accepted.
    4. The candidate solution is rejected.

    Each of these four outcomes is assigned a score :math:`s_j` (with
    :math:`j = 1,...,4`). After observing outcome :math:`j`, the weights of
    the selected destroy and repair operators :math:`d` and :math:`r` that
    were applied are updated as follows:

    .. math::

        \begin{align}
            \omega_d &= \theta \omega_d + (1 - \theta) s_j, \\
            \omega_r &= \theta \omega_r + (1 - \theta) s_j,
        \end{align}

    where :math:`0 \le \theta \le 1` (known as the *operator decay rate*)
    is a parameter.

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the weight
        updates when the candidate solution results in a new global best
        (idx 0), is better than the current solution (idx 1), the solution
        is accepted (idx 2), or rejected (idx 3).
    decay
        Operator decay parameter :math:`\theta \in [0, 1]`. This parameter is
        used to weigh the running performance of each operator.
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.
    """

    def __init__(
        self,
        scores: List[float],
        decay: float,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        super().__init__(num_destroy, num_repair, op_coupling)

        if any(score < 0 for score in scores):
            raise ValueError("Negative scores are not understood.")

        if len(scores) < 4:
            # More than four is OK because we only use the first four.
            raise ValueError(f"Expected four scores, found {len(scores)}")

        if not (0 <= decay <= 1):
            raise ValueError("decay outside [0, 1] not understood.")

        self._scores = scores
        self._d_weights = np.ones(num_destroy, dtype=float)
        self._r_weights = np.ones(num_repair, dtype=float)
        self._decay = decay

    @property
    def scores(self) -> List[float]:
        return self._scores

    @property
    def destroy_weights(self) -> np.ndarray:
        return self._d_weights

    @property
    def repair_weights(self) -> np.ndarray:
        return self._r_weights

    @property
    def decay(self) -> float:
        return self._decay

    def __call__(
        self, rnd_state: RandomState, best: State, curr: State
    ) -> Tuple[int, int]:
        """
        Selects a destroy and repair operator pair to apply in this iteration.
        The probability of an operator being selected is based on the operator
        weights: operators that frequently improve the current solution - and
        thus have higher weights - are selected with a higher probability.

        Parameters
        ----------
        rnd_state
            Random state object, to be used for random number generation.
        best
            The best solution state observed so far.
        current
            The current solution state.

        Returns
        -------
        A tuple of (d_idx, r_idx), which are indices into the destroy and
        repair operator lists, respectively.
        """

        def select(op_weights):
            probs = op_weights / np.sum(op_weights)
            return rnd_state.choice(range(len(op_weights)), p=probs)

        d_idx = select(self._d_weights)
        coupled_r_idcs = np.flatnonzero(self.op_coupling[d_idx])
        r_idx = coupled_r_idcs[select(self._r_weights[coupled_r_idcs])]

        return d_idx, r_idx

    def update(self, cand, d_idx, r_idx, outcome):
        self._d_weights[d_idx] *= self._decay
        self._d_weights[d_idx] += (1 - self._decay) * self._scores[outcome]

        self._r_weights[r_idx] *= self._decay
        self._r_weights[r_idx] += (1 - self._decay) * self._scores[outcome]



class VarObj:
    """Test solution state object with variable objective."""

    def __init__(self, obj: float):
        self.obj = obj

    def objective(self) -> float:
        return self.obj


class ContextualVarObj:
    """Test solution state object with variable objective and context."""

    def __init__(self, obj: float, context: list):
        self.obj = obj
        self.context = context

    def objective(self) -> float:
        return self.obj

    def get_context(self) -> list:
        return self.context


Sentinel = lambda: VarObj(0)  # noqa: E731
Zero = lambda: VarObj(0)  # noqa: E731
One = lambda: VarObj(1)  # noqa: E731
Two = lambda: VarObj(2)  # noqa: E731

ZeroWithZeroContext = lambda: ContextualVarObj(0, [0, 0, 0])  # noqa: E731
ZeroWithOneContext = lambda: ContextualVarObj(1, [1, 1, 1])  # noqa: E731



from typing import List

import numpy.random as rnd
from mabwiser.mab import LearningPolicy, NeighborhoodPolicy
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.Outcome import Outcome
from alns.select import MABSelector
from alns.select.MABSelector import arm2ops, ops2arm
from alns.tests.states import Zero, ZeroWithOneContext, ZeroWithZeroContext


@mark.parametrize(
    "destroy_idx, repair_idx",
    [
        (0, 0),
        (0, 1),
        (3, 3),
        (12, 7),
        (0, 14),
    ],
)
def test_arm_conversion(destroy_idx, repair_idx):
    expected = (destroy_idx, repair_idx)
    actual = arm2ops(ops2arm(destroy_idx, repair_idx))

    assert_equal(actual, expected)


def test_does_not_raise_on_valid_mab():
    policy = LearningPolicy.EpsilonGreedy(0.15)
    select = MABSelector([5, 0, 3, 0], 2, 1, policy)
    assert_equal(select.scores, [5, 0, 3, 0])
    assert_(len(select.mab.arms), 2)

    MABSelector([0, 0, 0, 0], 2, 1, policy, NeighborhoodPolicy.Radius(5))
    MABSelector(
        [1, 0, 0, 0], 2, 1, policy, NeighborhoodPolicy.Radius(5), 1234567
    )
    MABSelector([2, 1, 0, 0], 2, 1, policy, seed=1234567)


@mark.parametrize(
    "scores, learning_policy, num_destroy, num_repair",
    [
        (
            [5, 3, 2, -1],
            LearningPolicy.EpsilonGreedy(0.15),
            1,
            1,
        ),  # negative score
        (
            [5, 3, 2],
            LearningPolicy.EpsilonGreedy(0.15),
            1,
            1,
        ),  # len(score) < 4
    ],
)
def test_raises_invalid_arguments(
    scores: List[float],
    learning_policy: LearningPolicy,
    num_destroy: int,
    num_repair: int,
):
    with assert_raises(ValueError):
        MABSelector(scores, num_destroy, num_repair, learning_policy)


def test_call_with_only_one_operator_pair():
    # Only one operator pair, so the algorithm should select (0, 0).
    select = MABSelector(
        [2, 1, 1, 0], 1, 1, LearningPolicy.EpsilonGreedy(0.15)
    )
    state = rnd.RandomState()

    for _ in range(10):
        selected = select(state, Zero(), Zero())
        assert_equal(selected, (0, 0))


def test_mab_epsilon_greedy():
    state = rnd.RandomState()

    # epsilon=0 is equivalent to greedy selection
    select = MABSelector([2, 1, 1, 0], 2, 1, LearningPolicy.EpsilonGreedy(0.0))

    select.update(Zero(), 0, 0, outcome=Outcome.BETTER)
    selected = select(state, Zero(), Zero())
    for _ in range(10):
        selected = select(state, Zero(), Zero())
        assert_equal(selected, (0, 0))

    select.update(Zero(), 1, 0, outcome=Outcome.BEST)
    for _ in range(10):
        selected = select(state, Zero(), Zero())
        assert_equal(selected, (1, 0))


@mark.parametrize("alpha", [0.25, 0.5])
def test_mab_ucb1(alpha):
    state = rnd.RandomState()
    select = MABSelector([2, 1, 1, 0], 2, 1, LearningPolicy.UCB1(alpha))

    select.update(Zero(), 0, 0, outcome=Outcome.BEST)
    mab_select = select(state, Zero(), Zero())
    assert_equal(mab_select, (0, 0))

    select.update(Zero(), 0, 0, outcome=Outcome.REJECT)
    mab_select = select(state, Zero(), Zero())
    assert_equal(mab_select, (0, 0))


def test_contextual_mab_requires_context():
    select = MABSelector(
        [2, 1, 1, 0],
        2,
        1,
        LearningPolicy.LinGreedy(0),
    )
    # error: "Zero" state has no get_context method
    with assert_raises(AttributeError):
        select.update(Zero(), 0, 0, outcome=Outcome.BEST)


def text_contextual_mab_uses_context():
    state = rnd.RandomState()
    select = MABSelector(
        [2, 1, 1, 0],
        2,
        1,
        # epsilon=0 is equivalent to greedy
        LearningPolicy.LinGreedy(0),
    )

    select.update(ZeroWithZeroContext(), 0, 0, outcome=Outcome.REJECT)
    select.update(ZeroWithZeroContext(), 0, 0, outcome=Outcome.REJECT)
    select.update(ZeroWithZeroContext(), 1, 0, outcome=Outcome.BEST)

    select.update(ZeroWithOneContext(), 1, 0, outcome=Outcome.REJECT)
    select.update(ZeroWithOneContext(), 1, 0, outcome=Outcome.REJECT)
    select.update(ZeroWithOneContext(), 0, 0, outcome=Outcome.BEST)

    mab_select = select(state, ZeroWithZeroContext(), ZeroWithZeroContext())
    assert_equal(mab_select, (1, 0))

    mab_select = select(state, ZeroWithZeroContext(), ZeroWithZeroContext())
    assert_equal(mab_select, (0, 0))



from typing import List, Optional

import numpy as np

from alns.select.OperatorSelectionScheme import OperatorSelectionScheme


class AlphaUCB(OperatorSelectionScheme):
    """
    The :math:`\\alpha`-UCB (upper confidence bound) bandit scheme adapted
    from Hendel (2022).

    The action space :math:`A` is defined as each pair of (destroy, repair)
    operators that is allowed by the operator coupling matrix. The
    :math:`\\alpha`-UCB algorithm plays the following action in each iteration
    :math:`t`, computed as

    .. math::

        Q(t) = \\arg \\max_{a \\in A} \\left\\{ \\bar{r}_a (t - 1)
               + \\sqrt{\\frac{\\alpha \\ln(1 + t)}{T_a (t - 1)}} \\right\\},

    where :math:`T_a(t - 1)` is the number of times action :math:`a` has been
    played, and :math:`\\bar r_a(t - 1)` is the average reward of action
    :math:`a`, both in the first :math:`t - 1` iterations. The value that is
    maximised over the actions :math:`a \\in A` consists of the average reward
    term :math:`\\bar r_a(t - 1)` and an exploration bonus depending on
    :math:`t` and the number of times :math:`a` has been played.

    See :meth:`~alns.select.AlphaUCB.AlphaUCB.update` for details on how the
    average reward :math:`\\bar r_a` is updated.

    .. note::

        The average reward :math:`\\bar r_a(0)` of each action :math:`a \\in A`
        is initialised to 1. The ``scores`` list passed into the
        :math:`\\alpha`-UCB scheme should be 'reasonable' with respect to this
        default.

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the rewards when the
        candidate solution results in a new global best (idx 0), is better than
        the current solution (idx 1), the solution is accepted (idx 2), or
        rejected (idx 3).
    alpha
        The :math:`\\alpha \\in [0, 1]` parameter controls how much exploration
        is performed. Values of :math:`\\alpha`  near one result in much
        exploration, whereas values of :math:`\\alpha` nearer to zero result in
        more exploitation of good operator pairs. Typically,
        :math:`\\alpha \\le 0.1` is a good choice.
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    op_coupling
        Optional boolean matrix that indicates coupling between destroy and
        repair operators. Entry (i, j) is True if destroy operator i can be
        used together with repair operator j, and False otherwise.

    References
    ----------
    .. [1] Hendel, G. 2022. Adaptive large neighborhood search for mixed
           integer programming. *Mathematical Programming Computation* 14:
           185 - 221.
    """

    def __init__(
        self,
        scores: List[float],
        alpha: float,
        num_destroy: int,
        num_repair: int,
        op_coupling: Optional[np.ndarray] = None,
    ):
        super().__init__(num_destroy, num_repair, op_coupling)

        if not (0 <= alpha <= 1):
            raise ValueError(f"Alpha {alpha:} outside [0, 1] not understood.")

        if any(score < 0 for score in scores):
            raise ValueError("Negative scores are not understood.")

        if len(scores) < 4:
            # More than four is OK because we only use the first four.
            raise ValueError(f"Expected four scores, found {len(scores)}")

        self._scores = scores
        self._alpha = alpha

        self._avg_rewards = np.ones_like(self._op_coupling, dtype=float)
        self._times = np.zeros_like(self._op_coupling, dtype=int)
        self._iter = 0  # current iteration

    @property
    def scores(self) -> List[float]:
        return self._scores

    @property
    def alpha(self) -> float:
        return self._alpha

    def __call__(self, rnd, best, curr):
        """
        Returns the (destroy, repair) operator pair that maximises the average
        reward and exploration bonus.
        """
        action = np.argmax(self._values())
        return tuple(np.unravel_index(action, self.op_coupling.shape))

    def update(self, candidate, d_idx, r_idx, outcome):
        """
        Updates the average reward of the given destroy and repair operator
        combination ``(d_idx, r_idx)``.

        In particular, the reward of the action :math:`a` associated with this
        operator combination is updated as

        .. math::

            \\bar r_a (t) = \\frac{T_a(t - 1) \\bar r_a(t - 1)
                            + \\text{scores}[\\text{outcome}]}{T_a(t - 1) + 1},

        and :math:`T_a(t) = T_a(t - 1) + 1`.
        """
        # Update everything for the next iteration (t + 1)
        t_a = self._times[d_idx, r_idx]
        r = self._avg_rewards[d_idx, r_idx]
        avg_reward = (t_a * r + self.scores[outcome]) / (t_a + 1)

        self._avg_rewards[d_idx, r_idx] = avg_reward
        self._times[d_idx, r_idx] += 1
        self._iter += 1

    def _values(self):
        a = self.alpha
        t = self._iter

        value = self._avg_rewards
        explore_bonus = np.sqrt((a * np.log(1 + t)) / (self._times + 1))

        values = value + explore_bonus
        values[~self._op_coupling] = -1  # avoid selecting disallowed pairs

        return values



import sys
from importlib.metadata import version


def show_versions():
    """
    This function prints version information that is useful when filing bug
    reports.

    Examples
    --------
    Calling this function should print information like the following
    (dependency versions in your local installation will likely differ):

    >>> import alns
    >>> alns.show_versions()
    INSTALLED VERSIONS
    ------------------
          alns: 5.0.1
         numpy: 1.23.4
    matplotlib: 3.5.1
        Python: 3.9.9
    """
    python_version = ".".join(map(str, sys.version_info[:3]))

    print("INSTALLED VERSIONS")
    print("------------------")
    print(f"      alns: {version('alns')}")
    print(f"     numpy: {version('numpy')}")
    print(f"matplotlib: {version('matplotlib')}")
    print(f"    Python: {python_version}")
